#include <kos.h>
#include <time.h>
#include "font.h"
#include "edit.h"
#include "title.h"
#include "brkout.h"

#define vid_border_color(ARGS...) ;

#ifndef QUIT_TO_MENU
#define QUIT_TO_MENU 1
#endif

#define DEG2RAD (F_PI / 180.0f)

#define BALL_W 6
#define BALL_H 5
#define PADDLE_W 45
#define PADDLE_H 6

/* Ball velocity divisor: the inverse of how much push (relative to unit
 * velocity vector) to give the ball at each periodic speedup.  Increase to
 * make ball speed up less. */
#define BALL_VEL_DIV 5.0f
/* Cap max velocity to prevent ball passing through objects.  There should be
 * no need to change this, unless collision detection is adversely affected.
 * To adjust perceived maximum ball velocity, change TIMESCALE. */
#define MAX_VELOCITY 2.50f
/* Paddle speed -- if TIMESCALE was 1, this would be pixels moved per frame. */
#define PADDLE_SPD 6.0f

#define WALL_THICKNESS 5

/* Ratio of calculation passes to render passes.  Increasing TIMESCALE
 * decreases time granularity (allowing the ball to move faster and not pass
 * through objects).  Velocities such as ball and paddle velocity are scaled
 * with TIMESCALE so perceived velocity is not affected.  Perception of
 * unscaled velocities, such as MAX_VELOCITY, -is- affected. */
#define TIMESCALE 12

#define EXTRA_LIFE 25000
#define INITIAL_LIVES 3

static int play_game(void);
static void init_bricks(void);
static void render(void);
static int  handle_ball(void);
static void handle_paddle(void);
static void draw_text(void);
static void new_game(void);
static void next_level(void);
static void init_positions(void);
static void compile_headers(void);

int sprite_collision(int x1, int y1, int dx1, int dy1, int x2, int y2, int dx2, int dy2);
 
/* optimal random number function. */
unsigned long seed=456;
unsigned long init_seed = 456;
#define RNDC 1013904223
#define RNDM 1164525
#define RAND_MAX 0x7fffffff

int rand() {
	    seed = seed * RNDM + RNDC;
	    return seed & RAND_MAX;
}

void srand(unsigned long s) {
	seed = s;
}

int randnum(int limit) { 
	return (int)((double)rand() / ((double)RAND_MAX + 1) * limit);
}

/* Enable opaque and translucent polygons. */
static pvr_init_params_t pvr_params = {
	{ PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_16, PVR_BINSIZE_0, PVR_BINSIZE_0 },
	512 * 1024
};

level_t levels[NUM_LEVELS];
// brick_t bricks[BRICKS_X * BRICKS_Y];
color_t brick_colors[NUM_BRICKTYPES] = {
	{0, 0, 0},  /* unused */
	{255, 0, 0}, {255, 128, 0}, {255, 255, 0},
	/* {128, 255, 0}, */
	{0, 255, 0},
	/* {0, 255, 128}, */
	/* {0, 255, 255}, */
	{0, 192, 255}, 
	/* {0, 128, 255},  */
	{0, 0, 255},
	{128, 0, 255},
	{96, 96, 104},
	{229, 227, 211},
/*	{255, 0, 255} */
};

int brick_values[NUM_BRICKTYPES] = {
	100, 100, 100, 100, 100, 100, 100, 100, 500, 2500
};

typedef struct {
//	int brick;
	int x, y;
	brick_t type;
	float a;
} brick_fade_t;
brick_fade_t brick_fade = { 0, 0, 0, 0.0f };

pvr_poly_hdr_t op_poly_header;
uint8 maple_controller[4];
float paddle_x, paddle_y;
float ball_x, ball_y;
float ball_dx, ball_dy;
// int bricks_left;
int lives_left;
int level;
int last_level;

static int total_frames = 0;
static int score;
static int fourth_wall = 0;
static int cht_en = 0;

extern uint8 romdisk_boot[];
KOS_INIT_ROMDISK(romdisk_boot);

int main(void)
{
	int tret, i;

//	dbglog_set_level(DBG_DEBUG);

	maple_controller[0] = maple_first_controller();
	snd_stream_init(NULL);
    snd_sfx_load("/rd/click.wav");
	snd_sfx_load("/rd/activate.wav");
	snd_sfx_load("/rd/toggled.wav");
	snd_sfx_load("/rd/success.wav");
	snd_sfx_load("/rd/bounce.wav");
	snd_sfx_load("/rd/failure.wav");
	snd_sfx_load("/rd/disabled.wav");
	snd_sfx_load("/rd/cancel.wav");
	snd_sfx_load("/rd/savory.wav");

	pvr_init(&pvr_params);
	compile_headers();
	font_init();

	for (i = 0; i < NUM_LEVELS; i++) {
		levels[i].bricks = NULL;
	}
	last_level = 0;
	alloc_level(levels[0]);
	clear_level(levels[0]);

	load_levels(0); /* load included levels */

	cht_en = 0;

	for (;;) {
		tret = title_screen();
		switch (tret) {
			case MENU_PLAY:
				play_game();
				break;
			case MENU_EDIT:
				editor();
				break;
			case MENU_EMPTY:
				cht_en = 1;
				play_game();
				cht_en = 0;
				break;
			case MENU_QUIT:
				vid_border_color(0, 0, 0);
				kos_shutdown_all();
				if (QUIT_TO_MENU) { (*(void(**)())0x8c0000e0)(1); }
				return 0;
			default:
				printf("illegal menu choice %d\n", tret);
				break;
		}
	}

	return 0;
}

static int play_game(void)
{
	int t, i;
	cont_cond_t cond;
	cont_cond_t last_cond;
	int paddle_speed;
	int level_change = 0;
	int last_score;

	pvr_set_bg_color(0.06f, 0.06f, 0.10f);

	cont_get_cond(maple_controller[0], &last_cond);

	fourth_wall = 0;
	new_game();
	last_score = score;

	for (t = 0;;t++) {
		if (!(t % TIMESCALE)) render();

		if (handle_ball()) {
			if (lives_left <= 1) {
				for (i = 0; i < 60 * 3; i++) render(); 
				break;
			} else {
				lives_left--; 
				init_positions();
				for (i = 0; i < 30; i++) render();
			}
		} 
		handle_paddle();
		total_frames++; /* actually total calc passes */
		if (cont_get_cond(maple_controller[0], &cond)) {
			/* printf("Error getting controller %d status\n", 0); */
			continue;
		}
		if (levels[0].left == 0) { 
			ball_x = 1000;
			for (i = 0; i < 30; i++) render(); 
			next_level();
		}
		if (cond.rtrig) paddle_speed = 2; else paddle_speed = 1;
		if (!(cond.buttons & CONT_DPAD_LEFT)) paddle_x -= (PADDLE_SPD / TIMESCALE) * paddle_speed;
		if (!(cond.buttons & CONT_DPAD_RIGHT)) paddle_x += (PADDLE_SPD / TIMESCALE) * paddle_speed;
		paddle_x += ((cond.joyx - 128) / (128.0f / (PADDLE_SPD / TIMESCALE))) * paddle_speed;
		if (paddle_x < 0) paddle_x = 0;
		else if (paddle_x + PADDLE_W > FIELD_W) paddle_x = FIELD_W - PADDLE_W;

		if (BTN_PRESS(CONT_START)) {
			if (!(cond.buttons & CONT_B)) {
				break; 
			}
			else {
				do {
					last_cond = cond;
					render();
					cont_get_cond(maple_controller[0], &cond);
				} while (!BTN_PRESS(CONT_START));
			}
		}

		if (BTN_PRESS(CONT_B) && cht_en) { fourth_wall ^= 1; }
		if (BTN_PRESS(CONT_X) && cht_en) { 
			edit(level); 
			cont_get_cond(maple_controller[0], &last_cond);
			continue;
		}

		if (cond.ltrig && cond.rtrig) {
			if ((cond.buttons & CONT_Y)) {
				if (!level_change) {
					score = 0;
					last_score = 0;
					if (lives_left > INITIAL_LIVES) lives_left = INITIAL_LIVES;
					next_level();
				}
			} else {
				if (cht_en) {
					// hmm, seems redundant
					init_bricks();
					level++;
				}
			}
			level_change = 1;
		} else {
			level_change = 0;
		}
		last_cond = cond;
		lives_left += score / EXTRA_LIFE - last_score / EXTRA_LIFE;
		last_score = score;
	}
	return 0;
}



static void render(void) {
	vid_border_color(0, 0, 0);
	pvr_wait_ready();
	vid_border_color(255, 0, 0);
	pvr_scene_begin();

	pvr_list_begin(PVR_LIST_OP_POLY);
	draw_bricks(0);
	draw_walls();
	draw_playfield_bg();
	draw_ball(ball_x + FIELD_X,ball_y + FIELD_Y);
	draw_paddle(paddle_x + FIELD_X, paddle_y + FIELD_Y);
	draw_paddle(500, 250); /* for status display */
	if (fourth_wall) draw_paddle(535, 445);
	pvr_list_finish();

	pvr_list_begin(PVR_LIST_TR_POLY);
	pvr_prim(&trans_header, sizeof(trans_header));
	if (brick_fade.type > 0) {
		assert(brick_fade.a >= 0.0f && brick_fade.type < NUM_BRICKTYPES);
		if (brick_fade.a <= 1.0f) {
			draw_brick(brick_fade.x * BRICK_W + FIELD_X, brick_fade.y * BRICK_H + FIELD_Y, brick_fade.a, brick_colors[brick_fade.type].r,brick_colors[brick_fade.type].g,brick_colors[brick_fade.type].b);
		} else {
			draw_brick(brick_fade.x * BRICK_W + FIELD_X, brick_fade.y * BRICK_H + FIELD_Y, 1.0f, brick_colors[brick_fade.type].r,brick_colors[brick_fade.type].g,brick_colors[brick_fade.type].b);
			draw_brick(brick_fade.x * BRICK_W + FIELD_X, brick_fade.y * BRICK_H + FIELD_Y, 0.7f, 256, 256, 256);
		}

		brick_fade.a -= 0.1f;
		if (brick_fade.a < 0.0f) { brick_fade.type = 0; }
	}

	draw_text();
	pvr_list_finish();

	pvr_scene_finish();
	vid_border_color(0, 255, 0);
}

static void compile_headers(void) {
	pvr_poly_cxt_t tmp;
    pvr_poly_cxt_col(&tmp, PVR_LIST_OP_POLY);
    pvr_poly_compile(&op_poly_header, &tmp);
    pvr_poly_cxt_col(&tmp, PVR_LIST_TR_POLY);
    pvr_poly_compile(&trans_header, &tmp);
}

static void new_game(void) {
    srand((int)time(NULL));
	level = 0;
	score = 0;
	brick_fade.type = 0;
	lives_left = INITIAL_LIVES;
	next_level();
}

static void next_level(void) {
	int i;
	level++;
	init_bricks();
	init_positions();
	for (i = 0; i < 30; i++) render();
}

static void init_positions(void) {
	paddle_x = 200, paddle_y = FIELD_H - 20;
	ball_x = paddle_x, ball_y = paddle_y - 20.0f;
	ball_dx = 2.0f / TIMESCALE, ball_dy = 3.0f / TIMESCALE;
	/* Reset total_frames each time we init ball positions
	 * if you want each game to be identical.  Otherwise speedups
	 * will occur at different times. */
	/* total_frames = 1; */
}

	
static void init_bricks(void) {
	int i;
	int b;

	/* Generate random levels for levels greater than last_level. */
	if (level <= last_level) {
		copy_level(levels[0], levels[level]);
	} else {
		levels[0].left = 0;

		for (i = 0; i < BRICKS_X * BRICKS_Y; i++) {
			b = randnum(NUM_BRICKTYPES * 8);
			b -= NUM_BRICKTYPES * 7;
			if (b > 0) { 
				levels[0].left++;
				levels[0].bricks[i] = b;
			} else {
				levels[0].bricks[i] = 0;
			}
		}
	}

}

void draw_bricks(int lvl) {
	int i, j;
	brick_t brick;

	if (lvl > last_level) {
		printf("BUG: draw_bricks asked to draw illegal level %d > last_level %d\n", lvl, last_level);
		return; /* FIXME could hit the gcc return bug here--be careful */
	}

	pvr_prim(&op_poly_header, sizeof(op_poly_header));

	for (j = 0; j < BRICKS_Y; j++) {
		for (i = 0; i < BRICKS_X; i++) {
			brick = levels[lvl].bricks[j * BRICKS_X + i];
			if ((unsigned int)brick >= NUM_BRICKTYPES) {
				printf("BUG: brick at (%d, %d) is %d >= NUM_BRICKTYPES (%d), zeroing to protect!\n", i, j, brick, NUM_BRICKTYPES);
				brick = levels[lvl].bricks[j * BRICKS_X + i] = 0;
				levels[lvl].left--;
			}
			if (brick) {
				draw_brick(i * BRICK_W + FIELD_X, j * BRICK_H + FIELD_Y, 1.0f, brick_colors[brick].r,brick_colors[brick].g,brick_colors[brick].b);
			}
		}
	}
}

void draw_brick(int x, int y, float a, int r, int g, int b) {
	pvr_vertex_t vert;

	float z = 100.0f;
	int dx = BRICK_W, dy = BRICK_H;

	dx--; dy--;
	
	vert.flags = PVR_CMD_VERTEX;
	vert.x = x;
	vert.y = y + dy;
	vert.z = z;
	vert.u = vert.v = 0.0f;
	vert.argb = PVR_PACK_COLOR(a, r / 256.0f, g / 256.0f, b / 256.0f);
	vert.oargb = 0;
	pvr_prim(&vert, sizeof(vert));
	
	vert.y = y;
	pvr_prim(&vert, sizeof(vert));
	
	vert.x = x + dx;
	vert.y = y + dy;
	pvr_prim(&vert, sizeof(vert));

	vert.flags = PVR_CMD_VERTEX_EOL;
	vert.x = x + dx;
	vert.y = y;
	vert.argb = PVR_PACK_COLOR(a, r / 256.0f / 5, g / 256.0f / 5, b / 256.0f / 5);
	pvr_prim(&vert, sizeof(vert));

}

void draw_ball(float x, float y) {
	float z = 100.0f;
	int dx = BALL_W, dy = BALL_H;
	float r = 200 / 256.0f, g = 200 / 256.0f, b = 200 / 256.0f;

	draw_poly_box(x, y, dx, dy, z, 1.0f, r, g, b);

}

static int handle_ball(void) {
	static float last_x, last_y;
	static int ignore_paddle_collision = 0;

	int brx, bry, lbry, brick;
	int corners[4][2] = { { 0, 0 }, { BALL_W - 1, 0 }, { 0, BALL_H - 1 }, { BALL_W - 1, BALL_H - 1 } };
	int i;
	brick_t type;
	float vlen;
	float angle, angle_ydir;

	/* FIXME We can change ball direction 3 times during this function */

	if (ball_x < 0) {
		ball_dx = -ball_dx;
		snd_sfx_play(SAMPLE_BOUNCE, 150, 0);
	} else if (ball_x + BALL_W > FIELD_W) { 
		ball_dx = -ball_dx;
		snd_sfx_play(SAMPLE_BOUNCE, 150, 255);
	}
		
	if (ball_y < 0) {  
		ball_dy = -ball_dy;
		snd_sfx_play(SAMPLE_BOUNCE, 110, 128);
	}
	if (ball_y + BALL_H > FIELD_H) { 
		if (fourth_wall) { 
			ball_dy = -ball_dy;  
			snd_sfx_play(SAMPLE_BOUNCE, 110, 128);
		}
		else return 1;
	}

	for (i = 0; i < 4; i++) {
		brx = ( ball_x+corners[i][0] ) / BRICK_W;
		bry = ( ball_y+corners[i][1] ) / BRICK_H;
		if (brx >= 0 && brx < BRICKS_X && bry >= 0 && bry < BRICKS_Y) {
			brick = brx + bry * BRICKS_X;
			type = levels[0].bricks[brick];
			if (type) {
				levels[0].bricks[brick] = 0;
				score += brick_values[type];
				levels[0].left--;
				brick_fade.type = type;
				brick_fade.x = brx;
				brick_fade.y = bry;
				brick_fade.a = 1.2f;
				/* Rudimentary determination of which side we hit.
				 * If the last ball position was in the same row,
				 * reflect in x direction.  Otherwise, reflect y. */
				lbry = (last_y + corners[i][1]) / BRICK_H;
				if (bry == lbry) {
					ball_dx = -ball_dx;
				} else {
					ball_dy = -ball_dy;
				}
				snd_sfx_play(SAMPLE_EDITSEL_SELECT, 160, 128);
				break;
			}
		}
	}

	/* Collide with paddle.  Collision code depends on last frame having no collision,
	 * since we don't interpolate.  Ball incidence angle usually != reflection angle,
	 * so we can't guarantee we moved out of the paddle after a collision.
	 * Therefore, ignore further collisions for a couple calls. */
#define IGNORE_PADDLE_COLLISION 2

	if (sprite_collision(ball_x, ball_y, BALL_W, BALL_H, paddle_x, paddle_y, PADDLE_W, PADDLE_H)
			&& ignore_paddle_collision <= 0) {
		snd_sfx_play(SAMPLE_BOUNCE, 160, 128);
		for (i = 0; i < 4; i++) {
			/* Test every corner for paddle collision.  Then, check its previous position
			 * to see how to reflect it.  In some cases the calculation fails; when
			 * the northeast corner on a southeast-moving ball touches the northwest
			 * corner of the paddle, the ball will bounce up instead of left.  The 
			 * northeast corner shouldn't be tested in this case. */
			/* Direction is also wrong when we hit extreme edge of paddle (bounces
			 * backward), due to (I assume) either double hit or off-by-one error. */
			/* I've seen middle of paddle bounce backward as well ;) */
			if (ball_x + corners[i][0] >= paddle_x &&
					ball_x + corners[i][0] < paddle_x + PADDLE_W &&
					ball_y + corners[i][1] >= paddle_y &&
					ball_y + corners[i][1] < paddle_y + PADDLE_H) {
				/* This corner is in the paddle */
				if (last_y + corners[i][1] >= paddle_y && last_y + corners[i][1] < paddle_y + PADDLE_H) {
					/* Last position of this corner was at same vertical pos as paddle */
					ball_dx = -ball_dx;
					/* printf("* x change: corner %d @ (%f,%f) last (%f,%f) paddle (%f,%f)-(%f,%f)\n", i, ball_x + corners[i][0], ball_y + corners[i][1], last_x + corners[i][0], last_y + corners[i][1], paddle_x, paddle_y, paddle_x + PADDLE_W, paddle_y + PADDLE_H ); */
				} else {
					ball_dy = -ball_dy;
					vlen = fsqrt(ball_dy * ball_dy + ball_dx * ball_dx);
					angle_ydir = (ball_dy < 0) ? 180.0f: 0;

					/* Check where we hit the paddle, then create a unit vector with
					 * hardcoded bounce angle and scale it to the old velocity, then assign
					 * it to the ball velocity vector. */
					/* For positive horizontal motion, bounce off at decreasing angles as
					 * we move further to the right on the paddle. */
					/* I tried adding 90deg to the angle for negative motion.  The mirrors
					 * the paddle angles effectively (an 85 degree bounce angle becomes 15
					 * in the opposite direction at the same point on the paddle) but isn't
					 * practical, because large y components are acceptable while large x
					 * components are not (this is a vertical game!) For now I'm duplicating code :( */
					/* We add 180deg to the angle if dy is negative b/c sin(x) = -sin(x+180) */

#define ANGLE_OFFSET 180.0f

					if (ball_dx >= 0) {

						if (ball_x + BALL_W / 2 < paddle_x + PADDLE_W / 8) {
							angle = 85.0f;
						} else if (ball_x + BALL_W / 2 < paddle_x + PADDLE_W / 4) {
							angle = 70.0f;
						} else if (ball_x + BALL_W / 2 < paddle_x + PADDLE_W * 3 / 4) {
							angle = 57.0f;
						} else if (ball_x + BALL_W / 2 < paddle_x + PADDLE_W * 7 / 8) {
							angle = 35.0f;
						} else {
							angle = 20.0f;
						}
						ball_dx = vlen * fcos((angle) * DEG2RAD);
						ball_dy = vlen * fsin((angle + angle_ydir) * DEG2RAD);
					} else {
						if (ball_x + BALL_W / 2 < paddle_x + PADDLE_W / 8) {
							angle = 20.0f;
						} else if (ball_x + BALL_W / 2 < paddle_x + PADDLE_W / 4) {
							angle = 35.0f;
						} else if (ball_x + BALL_W / 2 < paddle_x + PADDLE_W * 3 / 4) {
							angle = 57.0f;
						} else if (ball_x + BALL_W / 2 < paddle_x + PADDLE_W * 7 / 8) {
							angle = 70.0f;
						} else {
							angle = 85.0f;
						}
						ball_dx = vlen * fcos( (angle + ANGLE_OFFSET) * DEG2RAD );
						ball_dy = vlen * fsin( (angle + angle_ydir) * DEG2RAD );
					}
				}
				ignore_paddle_collision = IGNORE_PADDLE_COLLISION;
				break; /* don't check any more corners */
			}
		}
		/* This is possible if ball is bigger than the side of the paddle it hit. */
		/* Actually, it always happens (because we break out of the for loop without setting
		 * a flag.  So we don't test for it right now. */
		/* printf("BUG: ball collided with paddle, but no corner is inside paddle.\n"); */
	} else {
		ignore_paddle_collision--;
	}

	if (total_frames % (300 * TIMESCALE) == 0) {
		/* Derive ball's velocity from its x and y components,
		 * then add 1/(5*timescale)th of a "velocity unit" ;) */
		vlen = fsqrt(ball_dy * ball_dy + ball_dx * ball_dx);
		/* Cap max velocity to prevent ball from passing thru objects.
		 * We won't catch this until one iteration after it occurs,
		 * but it won't matter in practice. */
		if (vlen <= MAX_VELOCITY) {
			ball_dx += ball_dx / vlen / (BALL_VEL_DIV * TIMESCALE);
			ball_dy += ball_dy / vlen / (BALL_VEL_DIV * TIMESCALE);
		}
	}

	last_x = ball_x; last_y = ball_y;
	ball_x += ball_dx; ball_y += ball_dy;
	return 0;
}

void handle_paddle(void) {
}

void draw_paddle(float x, float y) {
	float z = 100.0f;
	int dx = PADDLE_W, dy = PADDLE_H;
	float r = 100 / 256.0f, g = 100 / 256.0f, b = 150 / 256.0f;
	float a = 1.0f;

	draw_poly_box(x, y, dx, dy, z, a, r, g, b);
}

void draw_poly_box(float x, float y, float dx, float dy, float z, float a, float r, float g, float b) {
	pvr_vertex_t vert;

	vert.flags = PVR_CMD_VERTEX;
	vert.x = x;
	vert.y = y + dy;
	vert.z = z;
	vert.u = 0;
	vert.v = 1.0f;
	vert.argb = PVR_PACK_COLOR(a, r, g, b);
	vert.oargb = 0;
	pvr_prim(&vert, sizeof(vert));
	
	vert.y = y;
	vert.v = 0.0f;
	pvr_prim(&vert, sizeof(vert));
	
	vert.x = x + dx;
	vert.y = y + dy;
	vert.u = 1.0f;
	vert.v = 1.0f;
	pvr_prim(&vert, sizeof(vert));

	vert.flags = PVR_CMD_VERTEX_EOL;
	vert.y = y;
	vert.v = 0.0f;
	pvr_prim(&vert, sizeof(vert));
}

#define WALLS 3
static int wall_coords[WALLS][4] = { 
	{ FIELD_X - WALL_THICKNESS - 1, FIELD_Y - 1, WALL_THICKNESS, FIELD_H },
	{ FIELD_X + FIELD_W + 1, FIELD_Y - 1, WALL_THICKNESS, FIELD_H },
	{ FIELD_X - WALL_THICKNESS - 1, FIELD_Y - WALL_THICKNESS - 1, FIELD_W + 2 * WALL_THICKNESS + 2, WALL_THICKNESS }
};

void draw_walls(void) {
	float z = 100.0f;
	float r = 100 / 256.0f, g = 100 / 256.0f, b = 150 / 256.0f;
	float a = 1.0f;
	int i;

	for (i = 0; i < WALLS; i++)
		draw_poly_box(wall_coords[i][0], wall_coords[i][1], wall_coords[i][2], wall_coords[i][3],
				z, a, r, g, b);
}

void draw_playfield_bg(void) {
		float z = 50.0f;

		draw_poly_box(FIELD_X, FIELD_Y, FIELD_W, FIELD_H, z, 1.0f, 0, 0, 0);
}

void draw_cross(float x, float y, float dx, float dy) {
	float z = 300.0f;
	float size = 2.0f;
	float r = 1.0f, g = 1.0f, b = 1.0f;
	draw_poly_box(x - dx / 2, y - size / 2, dx, size, z, 1.0f, r, g, b);
	draw_poly_box(x - size / 2, y - dy / 2, size, dy, z, 1.0f, r, g, b);
}

#define abs(x) ((x) >= 0 ? (x) : (-x))

static void draw_text(void) {
	float vlen = fsqrt(ball_dx*ball_dx + ball_dy*ball_dy);
	pvr_prim(&font_header, sizeof(font_header));
	font_draw_strf(488, 50, 200, 1.0f, 1.0f, 1.0f, 1.0f, "~ brkout ~");
	font_draw_strf(520, 100, 200, 1.0f, 150 / 256.0f, 150 / 256.0f, 200 / 256.0f, "level");
	font_draw_strf(520, 130, 200, 1.0f, 1.0f, 1.0f, 1.0f, "%5d", level);
	font_draw_strf(520, 160, 200, 1.0f, 150 / 256.0f, 150 / 256.0f, 200 / 256.0f, "score");
	font_draw_strf(500, 190, 200, 1.0f, 1.0f, 1.0f, 1.0f, "%08d", score);
	font_draw_strf(500 + PADDLE_W + 15, 240, 200, 1.0f, 1.0f, 1.0f, 1.0f, "x");
	font_draw_strf(500 + 13 * 6, 240, 200, 1.0f, 1.0f, 1.0f, 1.0f, "%2d", lives_left - 1);
	font_draw_strf(490, 310, 200, 1.0f, 1.0f, 1.0f, 1.0f, "left %04d", levels[0].left);
	font_draw_strf(500, 350, 200, 1.0f, 1.0f, 1.0f, 1.0f, "dx %+1.2f", ball_dx);
	font_draw_strf(500, 380, 200, 1.0f, 1.0f, 1.0f, 1.0f, "dy %+1.2f", ball_dy);
	font_draw_strf(488, 410, 200, 1.0f, 1.0f, 1.0f, 1.0f, "vel %+1.2f", vlen);
	/* Unit vector -- describes the angle */
	/* font_draw_strf(480, 438, 200, 1.0f, 1.0f, 1.0f, 1.0f, "un %1.2f/%1.2f", abs(ball_dx / vlen), abs(ball_dy / vlen)); */
}


int sprite_collision(int x1, int y1, int dx1, int dy1, int x2, int y2, int dx2, int dy2)
{
	int right1 = x1 + dx1;
	int right2 = x2 + dx2;
	int bot1 = y1 + dy1;
	int bot2 = y2 + dy2;

	if (x2 < right1 && right2 > x1 && y2 < bot1 && bot2 > y1) return 1;
	return 0;
}

void field_to_screen(float *x, float *y) {
	*x = *x + FIELD_X;
	*y = *y + FIELD_Y;
}

int screen_to_field(float *x, float *y) {
	if (*x < FIELD_X || *x >= FIELD_X + FIELD_W || *y < FIELD_Y || *y >= FIELD_Y + FIELD_H)
		return 0;
	*x = *x - FIELD_X;
	*y = *y - FIELD_Y;
	return 1;
}

