#include <kos.h>
#include "font.h"
#include "edit.h"
#include "brkout.h"
#include <zlib.h>
#include <stdlib.h>

#define CROSS_XSPD 4.0f
#define CROSS_YSPD 3.0f

#define SELECTOR_X (FIELD_X + FIELD_W + 23)
#define SELECTOR_Y (FIELD_Y + 4)
#define ITEM_CHARS 3
#define ITEM_W (FONT_BOX_W * ITEM_CHARS)
#define ITEM_H (FONT_HEIGHT)
#define ITEMGAP_W 5.0f
#define ITEMGAP_H 2.0f
#define SELBOX_W (ITEMGAP_W + ITEM_W + ITEMGAP_W)
#define SELBOX_H (ITEMGAP_H + ITEM_H + ITEMGAP_H)
#define ITEMS_X 3
#define ITEMS_Y 10

#define MAX_SLOTS 99

#define dprintf(ARGS...) dbglog(DBG_DEBUG, ARGS)

typedef struct {
	float x, y, w, h, z, t;
	float a, r, g, b;
	float tr, tg, tb, ta;
	float xpad, ypad, yspacing;
	char *text;
} text_box_t;

struct menu_t;

typedef struct {
	char *text;
	enum { ENABLED, DISABLED, INVISIBLE } status;
	void (*activate)(void);
	void (*inc_dec)(struct menu_t *, int item, int dir);
} menu_item_t;

typedef struct {
	float x, y, w, h, z, t;
	float a, r, g, b;
	float ta, tr, tg, tb;
	float ba, br, bg, bb;
	float xpad, ypad, yspacing;
	float xpad_bar, ypad_bar;
	float bar_h;
	char *text;
	int selected;
	int nitems;
	menu_item_t *items;
} menu_t;

typedef struct {
	float ta, tr, tg, tb;  /* normal item color */
	float ha, hr, hg, hb;  /* highlighted item color */
	float sa[3], sr[3], sg[3], sb[3];  /* select bar color */
	text_box_t *box[3];
	int *hilite_list, nhilite;
 /* Disable: no items; items in hilite list; items /not/ in hilite list */
	enum { NONE, HILITE, NORMAL } disable;
	int sel[3];
	int active_sel;
	int top_row;
	unsigned int min, max;
	menu_t *mx, *my;
} selector_t;

static void edit_render(void);
static void edit_text(void);
static void draw_outline(float x, float y, float dx, float dy, float z, float width, float a, float r, float g, float b);
static int edit_select(int lvl);
static void edit_select_render(void);
static int save_levels(unsigned int slot);
static uint8 *load_vmu_icon(const char *path, uint16 *pal);
static void draw_text_box(text_box_t *tbox);
static void draw_menu(menu_t *menu);
static int edit_menu(menu_t *menu);
static void menu_vmu_save(void);
static void menu_vmu_load(void);
static void menu_vmu_port(void);
static void menu_vmu_port_incdec(struct menu_t *menu, int item, int dir);
static void menu_special_generic(void);
static int edit_select_handle(void);
static int list_search(int *list, int f, int n);
static int list_search_geq(int *list, int f, int n);
static inline void list_sort(int *list, int n);
static int scan_slots(void);
void draw_poly_tri(pvr_vertex_t *v1, pvr_vertex_t *v2, pvr_vertex_t *v3);
void draw_poly_tri_isosceles(float x, float y, float dx, float dy, float z, float a, float r, float g, float b);

static float cx, cy;
static int brx, bry;
static int cur_brick;
static int cur_level;
static int cur_slot = -1;
static level_t tmp_level;


pvr_poly_hdr_t trans_header;

static text_box_t text_box_help_save = {
	SELECTOR_X - 5, SELECTOR_Y + SELBOX_H * ITEMS_Y + 20, SELBOX_W * ITEMS_X + 10, 102, 600, 5,
	0.67f, 0.39f, 0.39f, 0.585f,
	1, 1, 1, 1,
	5, 5, 24,
	"A-Save\0 \0 \0B-Cancel\0"
};

static text_box_t text_box_help_load = {
	SELECTOR_X - 5, SELECTOR_Y + SELBOX_H * ITEMS_Y + 20, SELBOX_W * ITEMS_X + 10, 102, 600, 5,
	0.67f, 0.39f, 0.39f, 0.585f,
	1, 1, 1, 1,
	5, 5, 24,
	"A-Load\0 \0 \0B-Cancel\0"
};

static text_box_t text_box_help_generic = {
	SELECTOR_X - 5, SELECTOR_Y + SELBOX_H * ITEMS_Y + 20, SELBOX_W * ITEMS_X + 10, 102, 600, 5,
	0.67f, 0.39f, 0.39f, 0.585f,
	1, 1, 1, 1,
	5, 5, 24,
	"A-Select\0 \0 \0B-Cancel\0"
};

static text_box_t text_box_help_edit_sel = {
	SELECTOR_X - 5, SELECTOR_Y + SELBOX_H * ITEMS_Y + 20, SELBOX_W * ITEMS_X + 10, 102, 600, 5,
	0.67f, 0.39f, 0.39f, 0.585f,
	1, 1, 1, 1,
	5, 5, 24,
	"A-Edit\0X-Special\0Y-VMU\0B-Exit\0"
};

static text_box_t text_box_help_edit = {
	SELECTOR_X - 5, SELECTOR_Y + SELBOX_H * ITEMS_Y + 20, SELBOX_W * ITEMS_X + 10, 102, 600, 5,
	0.67f, 0.39f, 0.39f, 0.585f,
	1, 1, 1, 1,
	5, 5, 24,
	"A-Draw\0X-Erase\0Y-Revert\0B-Exit\0"
};

static text_box_t text_box_sel_type = {
	SELECTOR_X - (13 * 7) - 10, SELECTOR_Y - 5, 13 * 7, SELBOX_H + 5 + 2*ITEMGAP_H, 800, 5,
	0.90f, 0.70f, 0.20f, 0.20f,
	1, 1, 1, 1,
	5, 5 + ITEMGAP_H, 24,
	"\0"
};

static text_box_t text_box_sel2_type = {
	SELECTOR_X - (13 * 7) - 10, SELECTOR_Y - 5, 13 * 7, SELBOX_H + 5 + 2*ITEMGAP_H, 801, 5,
	0.90f, 0.20f, 0.70f, 0.20f,
	1, 1, 1, 1,
	5, 5 + ITEMGAP_H, 24,
	"\0"
};


#define MENU_VMU_ITEMS 3
static menu_item_t menu_vmu_items[MENU_VMU_ITEMS] = {
	{ "Save", ENABLED, menu_vmu_save, NULL },
	{ "Load", ENABLED, menu_vmu_load, NULL },
	{ "Port A1", DISABLED, menu_vmu_port, menu_vmu_port_incdec }
};

static menu_t menu_vmu = {
	200, 100, 200, 170, 700, 5,
	0.90f, 0.39f, 0.39f, 0.585f,
	1, 1, 1, 1,
	1.0f, 0.15f, 0.17f, 0.37f,
	5, 5, 35,
	5, 3,
	30,
	"VMU",
	1,
	MENU_VMU_ITEMS,
	menu_vmu_items
};

#define MENU_SPECIAL_ITEMS 7
static menu_item_t menu_special_items[MENU_SPECIAL_ITEMS] = {
	{ "Copy...", ENABLED, menu_special_generic, NULL },
	{ "Move...", ENABLED, menu_special_generic, NULL },
	{ "Swap...", ENABLED, menu_special_generic, NULL },
	{ "Insert", ENABLED, menu_special_generic, NULL },
	{ "Delete", ENABLED, menu_special_generic, NULL },
	{ "Delete All", ENABLED, menu_special_generic, NULL },
	{ "Clear", ENABLED, menu_special_generic, NULL }
};

enum { COPY, MOVE, SWAP, INSERT, DELETE, DELETE_ALL, CLEAR };


static menu_t menu_special = {
	200, 100, 200, 270, 700, 5,
	0.90f, 0.39f, 0.39f, 0.585f,
	1, 1, 1, 1,
	1.0f, 0.15f, 0.17f, 0.37f,
	5, 5, 30,
	5, 3,
	30,
	"Special",
	1,
	MENU_SPECIAL_ITEMS,
	menu_special_items
};

static selector_t selector_edit = {
	1.0f, 1.0f, 1.0f, 1.0f,
	1.0f, 0.4f, 0.4f, 0.8f,
	{ 1.0f,     0,     0 },
	{ 0.15f,    0,     0 },
	{ 0.17f,    0,     0 },
	{ 0.37f,    0,     0 },
	{ 0 },
	NULL, -1,
	NONE,
	{ -1, -1, -1 },
	-1,
	-1,
	0, 0,
	NULL, NULL
};

static selector_t selector_load = {
	1.0f, 0.4f, 0.4f, 0.8f,
	1.0f, 1.0f, 1.0f, 1.0f,
	{ 1.0f,   1.0f, 1.0f },
	{ 0.15f, 0.37f, 0.15f },
	{ 0.17f, 0.15f, 0.37f },
	{ 0.37f, 0.17f, 0.17f },
	{ 0 },
	NULL, -1,
	NONE,
	{ -1, -1, -1 },
	-1,
	-1,
	0, 0,
	NULL, NULL
};

static selector_t selector_save = {
	1.0f, 1.0f, 1.0f, 1.0f,
	1.0f, 0.4f, 0.4f, 0.8f,
	{ 1.0f,   1.0f, 1.0f },
	{ 0.15f, 0.37f, 0.15f },
	{ 0.17f, 0.15f, 0.37f },
	{ 0.37f, 0.17f, 0.17f },
	{ 0 },
	NULL, -1,
	NONE,
	{ -1, -1, -1 },
	-1,
	-1,
	0, 0,
	NULL, NULL
};

static selector_t selector_special = {
	1.0f, 1.0f, 1.0f, 1.0f,
	1.0f, 0.4f, 0.4f, 0.8f,
	{ 1.0f,   1.0f, 1.0f },
	{ 0.15f, 0.37f, 0.15f },
	{ 0.17f, 0.15f, 0.37f },
	{ 0.37f, 0.17f, 0.17f },
	{ 0 },
	NULL, -1,
	NONE,
	{ -1, -1, -1 },
	-1,
	-1,
	0, 0,
	NULL, NULL
};
	
static menu_t *cur_sel_menu;
static text_box_t *cur_help_text_box;
static selector_t *csel;

static void edit_render(void) {
	float x, y;

	vid_border_color(0,0,0);
	pvr_wait_ready();
	pvr_scene_begin();

	pvr_list_begin(PVR_LIST_OP_POLY);
	draw_bricks(0);
	draw_walls();
	draw_playfield_bg();
	x = ball_x; y = ball_y;
	if (x != -1 || y != -1) {
		field_to_screen(&x,&y);
		draw_ball(x, y);
		x = paddle_x; y = paddle_y;
		field_to_screen(&x,&y);
		draw_paddle(x, y);
	}
	draw_cross(cx,cy,28,21);
	draw_brick(520, 197, 1.0f, brick_colors[cur_brick].r, brick_colors[cur_brick].g, brick_colors[cur_brick].b );
	draw_poly_box(0,0,640,480,1,1.0f, .06f, .06f, .10f);
//	draw_paddle(500, 250); /* for status display */
	pvr_list_finish();

	pvr_list_begin(PVR_LIST_TR_POLY);
	pvr_prim(&trans_header, sizeof(trans_header));
	x = BRICK_W * brx;
	y = BRICK_H * bry;  /* brick to field */
	field_to_screen(&x, &y);
	draw_outline(x, y, BRICK_W, BRICK_H, 500.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f); // outline shadow brick
	draw_text_box(&text_box_help_edit);
//	draw_poly_box(x, y, BRICK_W, BRICK_H, 500.0f, 0.5f, 1.0f, 1.0f, 1.0f); // shadow brick
	edit_text();
	pvr_list_finish();

	pvr_scene_finish();
}

static int edit_menu(menu_t *menu) {
	int i;
	int item = 0;
	cont_cond_t cond, last_cond;

	cur_help_text_box = &text_box_help_generic;
	cur_sel_menu = menu;

	cont_get_cond(maple_controller[0], &last_cond);

	for (i = 0; ;i++) {
		menu->selected = item;

		/* This is spaghetti code */
		edit_select_render();

		if (cont_get_cond(maple_controller[0], &cond)) {
			continue;
		}

		if (BTN_PRESS(CONT_B)) {
			snd_sfx_play(SAMPLE_CANCEL, 160, 128);
			return -1;
		}
		if (BTN_PRESS(CONT_A)) { 
			snd_sfx_play(SAMPLE_ACTIVATE, 160, 128); /* maybe only if activate() called? */
			if (menu->items[item].activate) menu->items[item].activate(); 
			return item; 
		}
		if (BTN_PRESS(CONT_DPAD_RIGHT)) { if (menu->items[item].inc_dec) menu->items[item].inc_dec(menu, item, 1); }
		if (BTN_PRESS(CONT_DPAD_LEFT))  { if (menu->items[item].inc_dec) menu->items[item].inc_dec(menu, item, -1); }
		if (BTN_PRESS(CONT_DPAD_UP))    { item--; snd_sfx_play(SAMPLE_SELECT, 160, 128); }
		if (BTN_PRESS(CONT_DPAD_DOWN))  { item++; snd_sfx_play(SAMPLE_SELECT, 160, 128); }

		if (item < 0) item = menu->nitems - 1; if (item > menu->nitems - 1) item = 0;
		last_cond = cond;

	}

	return -2;
}

static int edit_select(int lvl) {
	static int last;
	int i;
	int ret;

	selector_edit.min = 1;
	selector_edit.sel[0] = lvl - 1;
	selector_edit.active_sel = 0;
	selector_edit.mx = &menu_special;
	selector_edit.my = &menu_vmu;
	selector_edit.hilite_list = &last;
	selector_edit.nhilite = 1;
	selector_edit.disable = NONE;

	for (;;) {
		last = last_level + 1;
		selector_edit.max = last;
		cur_sel_menu = NULL;
		cur_help_text_box = &text_box_help_edit_sel;
		csel = &selector_edit;

		ret = edit_select_handle();
		if (ret >= -1) return ret + selector_edit.min;
	}
}

/* Returns which index was selected, 0 ... n. 
 * Does -not- return value of item, perhaps it should. */

static int edit_select_handle(void) {
	int sel = csel->sel[csel->active_sel];
	int idx;
	int rpt_timer = 0;
	int moved = 0;

	assert(sel >= 0);
	assert(sel <= csel->max - csel->min);
	cont_cond_t cond, last_cond;
	cont_get_cond(maple_controller[0], &last_cond);

	for (;;) {
		/* FIXME  needs to be elsewhere */
		edit_select_render();
		if (cont_get_cond(maple_controller[0], &cond)) {
			/* printf("Error getting controller %d status\n", 0); */
			continue;
		}

		if (BTN_PRESS(CONT_B)) {
			snd_sfx_play(SAMPLE_CANCEL, 160, 128);
			return -1;
		}
		/* FIXME Need to figure out how to handle this at a higher level--
		 * will make code much cleaner */
		if (BTN_PRESS(CONT_Y)) { 
			if (csel->my) {
				snd_sfx_play(SAMPLE_ACTIVATE, 160, 128);
				edit_menu(csel->my); 
				return -2;
			}
		}
		if (BTN_PRESS(CONT_X)) { 
			if (csel->mx) {
				snd_sfx_play(SAMPLE_ACTIVATE, 160, 128);
				edit_menu(csel->mx); 
				return -3;
			}
		}
		if (BTN_PRESS(CONT_DPAD_RIGHT)) { sel++; moved = 1; }
		if (BTN_PRESS(CONT_DPAD_LEFT))  { sel--; moved = 1; }
		if (BTN_PRESS_REPEAT(CONT_DPAD_UP))   { 
			sel -= ITEMS_X; REPEAT_UPDATE(CONT_DPAD_UP); 
			moved = 1;
		}
		if (BTN_PRESS_REPEAT(CONT_DPAD_DOWN)) { 
			sel += ITEMS_X; REPEAT_UPDATE(CONT_DPAD_DOWN); 
			moved = 1;
		}

		if (cond.ltrig && !last_cond.ltrig) { sel = 0; moved = 1; }
		if (cond.rtrig && !last_cond.rtrig) { sel = csel->max - csel->min; moved = 1; }

		/* need a function to retrieve the number under selector, given sel */
		if (moved) {
			snd_sfx_play(SAMPLE_EDITSEL_SELECT, 160, 128);
			if (sel < 0) 
				sel = 0; 
			if (sel > csel->max - csel->min)
				sel = csel->max - csel->min; 
			csel->sel[csel->active_sel] = sel;
			moved = 0;
		}

		if (BTN_PRESS(CONT_A)) {
			if (csel->disable == NONE) {
				snd_sfx_play(SAMPLE_ACTIVATE, 160, 128);
				return sel;
			}
			if (csel->hilite_list) {
				idx = list_search(csel->hilite_list, sel + csel->min, csel->nhilite);
			} else idx = -1;
			if ((idx >= 0 && csel->disable == NORMAL) || 
				(idx <  0 && csel->disable == HILITE)) {
				snd_sfx_play(SAMPLE_ACTIVATE, 160, 128);
				return sel;
			}
			/* Otherwise, we selected a disabled item, so continue. */
			snd_sfx_play(SAMPLE_DISABLED, 160, 128);
		}

		last_cond = cond;
		rpt_timer--;

	}
}

static void edit_select_render(void) {
	int i, j;
	int selx[3], sely[3];
	int last_row;
	int num;
	int h;

	for (i = 0; i < 3; i++) {
		selx[i] = csel->sel[i] % ITEMS_X;
		sely[i] = csel->sel[i] / ITEMS_X;
	}

	if (csel->top_row == -1) csel->top_row = 0; /* FIXME temporary */

	if (csel->top_row > sely[csel->active_sel]) 
		csel->top_row = sely[csel->active_sel];
	else if (csel->top_row < sely[csel->active_sel] - ITEMS_Y + 1)
		csel->top_row = sely[csel->active_sel] - ITEMS_Y + 1;
	assert(csel->top_row >= 0); /* first row */
	last_row = (csel->max - csel->min) / ITEMS_X;
	assert(csel->top_row <= last_row);

	vid_border_color(0,0,0);
	pvr_wait_ready();
	pvr_scene_begin();

	pvr_list_begin(PVR_LIST_OP_POLY);

	/* Draw background */
	pvr_prim(&op_poly_header, sizeof(op_poly_header));
	draw_poly_box(0,0,640,480,1,1.0f, .06f, .06f, .10f);
	/* Draw selected level, if it's legal */
	if (selector_edit.sel[0] + selector_edit.min < last_level + 1 && selector_edit.sel[0] >= 0) { draw_bricks(selector_edit.sel[0] + selector_edit.min); }
	pvr_prim(&op_poly_header, sizeof(op_poly_header));
	draw_walls();
	draw_playfield_bg();

	pvr_list_finish();

	pvr_list_begin(PVR_LIST_TR_POLY);

	/* Draw level selector */
	pvr_prim(&trans_header, sizeof(trans_header));
	/* Yes, the outline and background overlap */
	/* FIXME need to move this to draw_outline_box */
	draw_poly_box(SELECTOR_X - 10, SELECTOR_Y - 10, SELBOX_W * ITEMS_X + 20, SELBOX_H * ITEMS_Y +20, 100, 0.7f, 0.0f, 0.0f, 0.0f);
	draw_outline(SELECTOR_X - 10, SELECTOR_Y - 10, SELBOX_W * ITEMS_X + 20, SELBOX_H * ITEMS_Y + 20, 500.0f, 5.0f, 1.0f, 0.39f, 0.39f, 0.585f);
	for (i = 0; i < 3; i++) {
		if (csel->sel[i] >= 0) {
			sely[i] -= csel->top_row;
			if (sely[i] >= 0 && sely[i] < ITEMS_Y) {
				draw_poly_box(SELECTOR_X + SELBOX_W * selx[i], SELECTOR_Y + SELBOX_H * sely[i], 
				              SELBOX_W, SELBOX_H, 150.0f, csel->sa[i], csel->sr[i], csel->sg[i], csel->sb[i]);
			}
		}
	}

	draw_text_box(cur_help_text_box);
	
	/* One menu can be associated with the selector.
	 * Null pointers are skipped.  Horrible idea, simple implementation. */
	draw_menu(cur_sel_menu);

	for (i = 0; i < 3; i++) {
		if (csel->box[i] && csel->sel[i] >= 0) {
			if (sely[i] < 0) 
				sely[i] = 0;
			if (sely[i] > ITEMS_Y - 1) 
				sely[i] = ITEMS_Y - 1;
			csel->box[i]->y = SELECTOR_Y - 5 + sely[i] * SELBOX_H;
			draw_text_box(csel->box[i]);
		}
	}

#define T_WIDTH 20.0f
#define T_HEIGHT 20.0f
	if (csel->top_row > 0) {
		pvr_prim(&trans_header, sizeof(trans_header));
		draw_poly_tri_isosceles(SELECTOR_X + SELBOX_W * ITEMS_X - 2.5f, SELECTOR_Y - 5.0f, 
			T_WIDTH, T_HEIGHT, 501.0f, 1.0f, 0.4f, 0.4f, 0.6f);
	}
	if (csel->top_row + ITEMS_Y <= last_row) {
		pvr_prim(&trans_header, sizeof(trans_header));
		draw_poly_tri_isosceles(SELECTOR_X + SELBOX_W * ITEMS_X - 2.5f, SELECTOR_Y + SELBOX_H * ITEMS_Y + 5.0f,
			T_WIDTH, -T_HEIGHT, 901.0f, 1.0f, 0.4f, 0.4f, 0.6f);
	}
#undef T_WIDTH
#undef T_HEIGHT

	num = csel->min + ITEMS_X * csel->top_row;
	/* The incremental searching algorithm below requires that the current
	 * element of the search list be >= to the searched-for number. */
	h = -1;
	if (csel->hilite_list) 
		h = list_search_geq(csel->hilite_list, num, csel->nhilite);
	pvr_prim(&font_header, sizeof(font_header));
	for (j = 0; j < ITEMS_Y; j++) {
		for (i = 0; i < ITEMS_X; i++) {
			if (num > csel->max) { j = ITEMS_Y; break; /* break out of both for loops */ }
			if (csel->hilite_list && h >= 0 && h < csel->nhilite && csel->hilite_list[h] == num) {
				font_draw_strf(SELECTOR_X + SELBOX_W * i + ITEMGAP_W, SELECTOR_Y + SELBOX_H * j + ITEMGAP_H, 200, csel->ha, csel->hr, csel->hg, csel->hb, "% 3d", num);
				h++;
			} else {
				font_draw_strf(SELECTOR_X + SELBOX_W * i + ITEMGAP_W, SELECTOR_Y + SELBOX_H * j + ITEMGAP_H, 200, csel->ta, csel->tr, csel->tg, csel->tb, "% 3d", num);
			}
			num++;
		}
	}
	
	pvr_list_finish();

	pvr_scene_finish();
}



void editor(void) {
	int lvl = 1;

	scan_slots();

	while ((lvl = edit_select(lvl))) {
		if (lvl > last_level) {
			if (lvl == last_level + 1) {
				if (lvl >= NUM_LEVELS) continue;
				last_level++;
				alloc_level(levels[lvl]);
				clear_level(levels[lvl]);
			} else {
				printf("BUG: editor() trying to edit randomized level\n");
				return;
			}
		}

		cur_level = lvl;
		copy_level(levels[0], levels[cur_level]); 
		/* Set ball x and y to -1 so that edit() does not register cross motion,
		 * preventing a brick from being placed when we select level with A. */
		/* edit_render checks the -1 value to determine if we should draw ball/paddle */
		/* Alternatively we could set a flag saying "i'm calling from editor not play_game" */
		ball_x = -1; ball_y = -1;
		edit(cur_level);
		copy_level(levels[cur_level], levels[0]);
	}

//	save_levels();
}

void edit(int lvl) {
	int i, j;
	float bx, by;
	int tbrx, tbry;
	int brick;
	float ocx, ocy;
	int rpt_timer = 0;
	cont_cond_t cond, last_cond;
	cx = ball_x; cy = ball_y;
	ocx = -1; ocy = -1;

	brx = 0; bry = 0; 	/* Actually, this should be calculated. */
	field_to_screen(&cx, &cy);
	cur_brick = 1;
	
	cur_level = lvl; /* FIXME unfortunate kludge since brkout.c can't access cur_level */

	/* Save a copy of this level, so we can discard changes */
	/* Only useful when entering edit mode in the middle of play (since we always have
	 * a valid copy when entering from the editor) */
	alloc_level(tmp_level);
	copy_level(tmp_level, levels[0]);

	cont_get_cond(maple_controller[0], &last_cond);

	for (i = 0; ;i++) {
		/* Adjust shadow brick location to cross position
		 * if it's onfield within the brick range. */
		/* I don't like this at the top of the function */
		if (ocx != cx || ocy != cy) {
			bx = cx; by = cy;
			if (screen_to_field(&bx, &by)) {
				/* this brick calculation should be a function */
				tbrx = (int)bx / BRICK_W;
				tbry = (int)by / BRICK_H;
				if (tbrx >= 0 && tbrx < BRICKS_X && tbry >= 0 && tbry < BRICKS_Y) {
					/* We're within the brick field */
					if (tbrx != brx || tbry != bry) {
						/* We've moved */
						brx = tbrx; bry = tbry;
						last_cond.buttons |= CONT_A | CONT_X;
					}
				}
			} else {
				/* off field */
			}
			ocx = cx; ocy = cy;
			/* Bleah, this is becoming hairy to have separate joystick logic. */
		}

		edit_render();

		if (cont_get_cond(maple_controller[0], &cond)) {
			/* printf("Error getting controller %d status\n", 0); */
			continue;
		}
		/* B: exit edit mode */
		if (BTN_PRESS(CONT_B)) { 
			snd_sfx_play(SAMPLE_CANCEL, 160, 128);
			break;
		}
		if (BTN_PRESS(CONT_Y)) {
				// clear_level(levels[0]);
				copy_level(levels[0], tmp_level); /* reset level */
		}
		/* A: toggle brick at shadow brick location */
		if (BTN_PRESS(CONT_A)) {
				brick = brx + bry * BRICKS_X;
				if (!levels[0].bricks[brick]) {
					levels[0].left++;
				}
				levels[0].bricks[brick] = cur_brick;
		}

		if (BTN_PRESS(CONT_X)) {
				brick = brx + bry * BRICKS_X;
				if (levels[0].bricks[brick]) { 
					levels[0].bricks[brick] = 0;
					levels[0].left--;
				}
		}

		if (cond.rtrig && !last_cond.rtrig) {
			cur_brick++;
			if (cur_brick >= NUM_BRICKTYPES) cur_brick = 1;
		}

		if (cond.ltrig && !last_cond.ltrig) {
			cur_brick--;
			if (cur_brick < 1) cur_brick = NUM_BRICKTYPES - 1;
		}

		/* Process analog stick */
		cx += ((cond.joyx - 128) / (128.0f / (CROSS_XSPD)));
		cy += ((cond.joyy - 128) / (128.0f / (CROSS_YSPD)));
		if (cx < 0) cx = 0; else if (cx > 639) cx = 639;
		if (cy < 0) cy = 0; else if (cy > 479) cx = 479;

		/* Process dpad, and move shadow brick */
		tbrx = brx; tbry = bry;
		if (BTN_PRESS_REPEAT(CONT_DPAD_LEFT))  { brx--; if (brx < 0) brx = 0; REPEAT_UPDATE(CONT_DPAD_LEFT); }
		if (BTN_PRESS_REPEAT(CONT_DPAD_RIGHT)) { brx++; if (brx >= BRICKS_X) brx = BRICKS_X - 1; REPEAT_UPDATE(CONT_DPAD_RIGHT); }
		if (BTN_PRESS_REPEAT(CONT_DPAD_UP))    { bry--; if (bry < 0) bry = 0; REPEAT_UPDATE(CONT_DPAD_UP); }
		if (BTN_PRESS_REPEAT(CONT_DPAD_DOWN))  { bry++; if (bry >= BRICKS_Y) bry = BRICKS_Y - 1; REPEAT_UPDATE(CONT_DPAD_DOWN); }
		if (tbrx != brx || tbry != bry) {
			/* brick location changed */
			cond.buttons |= CONT_A | CONT_X;
		}
		
		last_cond = cond;
		rpt_timer--;

	}

	free_level(tmp_level);

}

static void edit_text(void) {
	float vlen = fsqrt(ball_dx*ball_dx + ball_dy*ball_dy);
	pvr_prim(&font_header, sizeof(font_header));
	font_draw_strf(488, 50, 200, 1.0f, 1.0f, 1.0f, 1.0f, "~  edit  ~");
	font_draw_strf(520, 100, 200, 1.0f, 150 / 256.0f, 150 / 256.0f, 200 / 256.0f, "level");
	font_draw_strf(520, 130, 200, 1.0f, 1.0f, 1.0f, 1.0f, "%5d", cur_level);
	font_draw_strf(520, 160, 200, 1.0f, 150 / 256.0f, 150 / 256.0f, 200 / 256.0f, "brick");
	font_draw_strf(520 + 13 * 2, 190, 200, 1.0f, 1.0f, 1.0f, 1.0f, "%3d", cur_brick);
	font_draw_strf(494, 220, 200, 1.0f, 150/256.0f, 150/256.0f, 200/256.0f, "<- L R ->");
//	font_draw_strf(500 + PADDLE_W + 15, 240, 200, 1.0f, 1.0f, 1.0f, 1.0f, "x");
//	font_draw_strf(500 + 13 * 6, 240, 200, 1.0f, 1.0f, 1.0f, 1.0f, "%2d", lives_left - 1);
	font_draw_strf(490, 310, 200, 1.0f, 1.0f, 1.0f, 1.0f, "left %04d", levels[0].left);
//	font_draw_strf(500, 350, 200, 1.0f, 1.0f, 1.0f, 1.0f, "dx %+1.2f", ball_dx);
//	font_draw_strf(500, 380, 200, 1.0f, 1.0f, 1.0f, 1.0f, "dy %+1.2f", ball_dy);
//	font_draw_strf(488, 410, 200, 1.0f, 1.0f, 1.0f, 1.0f, "vel %+1.2f", vlen);
	/* Unit vector -- describes the angle */
	/* font_draw_strf(480, 438, 200, 1.0f, 1.0f, 1.0f, 1.0f, "un %1.2f/%1.2f", abs(ball_dx / vlen), abs(ball_dy / vlen)); */
}

static void draw_outline(float x, float y, float dx, float dy, float z, float width, float a, float r, float g, float b) {
	draw_poly_box(x, y, dx, width, z, a, r, g, b); //top
	draw_poly_box(x, y + dy - width, dx, width, z, a, r, g, b); //bottom
	draw_poly_box(x, y, width, dy, z, a, r, g, b);
	draw_poly_box(x + dx - width, y, width, dy, z, a, r, g, b);
}	

#ifndef SAVE_DIR
#  define SAVE_DIR "/vmu/a1"
#endif
#ifndef LOAD_DIR
#  define LOAD_DIR "/vmu/a1"
#endif
#define SAVE_FORMAT_VERSION 1
#define HEADER_SIZE (2 * sizeof(int))
#define LEVEL_HEADER_SIZE (4 * sizeof(int))
#define ZIP_HEADER_SIZE (2 * sizeof(int))
#undef LOAD_ORIGINAL_FORMAT

/* Format spec:
 * 640 bytes vmu header (128 header + 512 icon)
 * 3 bytes "BRK" + 1 byte format version (currently 1)
 * 4 bytes (int) uncompressed data size
 * followed by compressed data:
 *   2 bytes (short) no. of bricks horizontally (==bricks_x)
 *   2 bytes (short) no. of bricks vertically   (==bricks_y)
 *   4 bytes (int) number of levels
 *   then one or more:
 *     4 bytes (int) no. of bricks this level 
 *     12 bytes reserved
 *     bricks_x * bricks_y bytes level data, 1 char (brick_t) per brick
 */

/* returns -1 on failure, 0 on success */
int load_levels(unsigned int slot) {
	int lvl;
	int i;
	file_t f;
	char *buf = NULL, *pbuf = NULL;
	char *dbuf = NULL;
	size_t size;
	unsigned long dsize;
	vmu_hdr_t hdr;
	char icon[512];
	char fn[1024];
	char sn[4];
	short bx, by;
	char version;

	if (slot < 0 || slot > 999) {
		printf("slot %u out of range\n", slot);
		return -1;
	}

	if (slot == 0) {
		strcpy(fn, "/rd/brkout.000");
	} else {
		/* FIXME buffer overflow ... need snprintf */
		sprintf(sn, "%03u", slot);
		strcpy(fn, LOAD_DIR "/brkout.");
		strcat(fn, sn);
	}

	printf("loading levels from %s\n", fn);
	f = fs_open(fn, O_RDONLY);
	if (!f) return -1;
	size = fs_total(f);
	printf("size is %d bytes\n", size);
	if (!size) return -1;
	if (fs_read(f, &hdr, sizeof(vmu_hdr_t)) != sizeof(vmu_hdr_t)) {
		printf("failed to read vmu_hdr\n");
		fs_close(f);
		return -1;
	}

	/* There's no simple way to detect and skip over eyecatch/icon data
	 * without writing a vmu_pkg_parse function -- so we hardcode it for now */
	dprintf("icon count: %d\n", hdr.icon_cnt);
	for (i = 0; i < hdr.icon_cnt; i++) {
		/* VMU can't fs_seek! bleah! */
		fs_read(f, icon, 512);
	}
	size = hdr.data_len;
	printf("data size: %d bytes\n", size);
	if (size <= 0) {
		fs_close(f);
		return -1;
	}
	buf = (char *)malloc(size);

	pbuf = buf;
	if (fs_read(f, buf, size) != size) {
		printf("failed to read\n");
		fs_close(f);
		free(buf);
		return -1;
	}
	fs_close(f);

#ifndef LOAD_ORIGINAL_FORMAT
	if (strncmp(buf, "BRK", 3) != 0) {
		printf("bad header, load aborted.\n");
		free(pbuf);
		return -1;
	}
	buf += 3;
	version = *buf++;
	if (version != SAVE_FORMAT_VERSION) {
		printf("unrecognized file version %d\n", (int)version);
		free(pbuf);
		return -1;
	}
#endif

	dsize = (unsigned long) (*((int *)buf)++);
	dbuf = (char *)malloc(dsize);

	if (uncompress(dbuf, &dsize, buf, size) != Z_OK) {
		printf("uncompress failed\n");
		free(dbuf);
		free(pbuf);
		return -1;
	}

	printf("uncompress successful\n");

	free(pbuf);

#ifndef LOAD_ORIGINAL_FORMAT
	bx = *((short *)dbuf)++;
	by = *((short *)dbuf)++;
	if (bx != BRICKS_X || by != BRICKS_Y) {
		printf("file specified BRICKS_X: %d BRICKS_Y: %d, we need X %d Y %d\n", (int)bx, (int)by, BRICKS_X, BRICKS_Y);
		free(dbuf);
		return -1;
	}
#endif

	lvl = *((int *)dbuf)++;

	printf("number of levels: %d\n", lvl);
	if (lvl >= NUM_LEVELS) {
		printf("lvl > NUM_LEVELS, adjusting\n");
		lvl = NUM_LEVELS - 1; 
	}
	/* We're not cleaning up any corrupt data, so beware.  Mostly
	 * this would affect draw_bricks (which now has a num_bricktypes assert). */
	for (i = 1; i <= lvl; i++) {
		free_level(levels[i]); /* no-op for NULL pointers */
		alloc_level(levels[i]);
		levels[i].left = *((int *)dbuf)++;
#ifndef LOAD_ORIGINAL_FORMAT
		(int *)dbuf += 3; /* reserved */
#endif
		memcpy(levels[i].bricks, dbuf, LEVEL_SIZE);
		dbuf += LEVEL_SIZE;
		printf("read in level %d with left %d\n", i, levels[i].left);
	}
	for (i = lvl + 1; i <= last_level; i++) {
		/* free any higher levels (if we have fewer levels now) */
		free_level(levels[i]);
		levels[i].bricks = NULL;
		printf("freed level %d\n", i);
	}
	last_level = lvl;

	free(dbuf);

	return slot;
}

static int save_levels(unsigned int slot) {
	file_t f;
	int i;
	char *buf, *pbuf;
	char *cbuf;
	size_t size;
	unsigned long csize;
    vmu_pkg_t   pkg;
	uint8 *pkg_out;
	int     pkg_size;
	int reason;
	char fn[1024];
	char sn[4];

	if (slot < 1 || slot > 999) {
		printf("save slot %u out of range\n", slot); 
		return -1;
	}

	/* FIXME buffer overflow ... need snprintf */
	sprintf(sn, "%03u", slot);
	strcpy(fn, SAVE_DIR "/brkout.");
	strcat(fn, sn);

	printf("saving levels to %s\n", fn);
	
	size = HEADER_SIZE + (LEVEL_HEADER_SIZE + LEVEL_SIZE) * last_level;
	buf = malloc(size);
	printf("buf: %08x\n", (uint32)buf);
	printf("malloced %d\n", size);
	pbuf = buf;

	/* begin HEADER */
	*((short *)buf)++ = BRICKS_X;
	*((short *)buf)++ = BRICKS_Y;
	*((int *)buf)++ = last_level;
	/* end HEADER */

	printf("buf: %08x\n", (uint32)buf);
	for (i = 1; i <= last_level; i++) {
		printf("writing out level %d with left %d\n", i, levels[i].left);
		/* begin LEVEL_HEADER */
		*((int *)buf)++ = levels[i].left;
		*((int *)buf)++ = 0;
		*((int *)buf)++ = 0;
		*((int *)buf)++ = 0;
		/* end LEVEL_HEADER */

		printf("buf: %08x ", (uint32)buf);
		memcpy(buf, levels[i].bricks, LEVEL_SIZE);
		buf += LEVEL_SIZE;
		printf("wrote out level %d with left %d\n", i, levels[i].left);
	}
	csize = (int)(size * 1.1f) + 12;
	cbuf = malloc(csize + ZIP_HEADER_SIZE);
	printf("cbuf: %08x csize: %d\n",(uint32)cbuf, csize);
	/* skip over header */
	cbuf += ZIP_HEADER_SIZE;
	if ((reason = compress(cbuf, &csize, pbuf, (unsigned long)size)) != Z_OK) {
		printf("failed to compress, reason %d\n", reason);
		free(cbuf);
		free(pbuf);
		return -1;
	}
	/* go to beginning of header */ 
	cbuf -= ZIP_HEADER_SIZE;
	/* begin ZIP_HEADER */
	*cbuf++ = 'B'; *cbuf++ = 'R'; *cbuf++ = 'K'; *cbuf++ = (char)SAVE_FORMAT_VERSION;
	*((int *)cbuf)++ = size;
	cbuf -= ZIP_HEADER_SIZE;
	/* end ZIP_HEADER */

	strcpy(pkg.desc_short, "brkout");
	strcpy(pkg.desc_long, "brkout level data ");
	strcat(pkg.desc_long, sn);
	strcpy(pkg.app_id, "brkout");
	pkg.icon_cnt = 1;
	pkg.icon_anim_speed = 0;
	pkg.icon_data = load_vmu_icon("/rd/brkout-icon.bmp", pkg.icon_pal);
	pkg.eyecatch_type = VMUPKG_EC_NONE;
	pkg.data_len = csize + HEADER_SIZE;
	pkg.data = cbuf;

	if (vmu_pkg_build(&pkg, &pkg_out, &pkg_size) < 0) {
		printf("failed to build package");
	} else {
		printf("package size: %d; pkg_out %08x\n", pkg_size, (uint32)pkg_out);
		fs_unlink(fn);
		printf("unlinked %s\n", fn);
		f = fs_open(fn, O_WRONLY);
		printf("opened file\n");
		if (!f) return -1;
		if (fs_write(f, pkg_out, pkg_size) != pkg_size) {
			printf("** FAILED to write\n");
		}
		fs_close(f);
	}
	free(cbuf);
	free(pbuf);
	free(pkg.icon_data);
	printf("number of levels: %d\n", last_level);
	return slot;
}

/* Load a palettized bmp. */
static uint8 *load_vmu_icon(const char *path, uint16 *pal) {
	file_t f;
	unsigned short bpp;
	int ncolors; 
	unsigned short magic;
	unsigned int height, width, clr_used;
	unsigned char r, g, b;
	int i;
	int size, rsize;
	uint8 *buf;

	dprintf("opening icon file %s\n", path);
	f = fs_open(path, O_RDONLY);
	if (!f) return NULL;
	fs_read(f, &magic, 2);
	if (magic != 19778) {
		fs_close(f);
		printf("%s: not a BMP file (%u)\n", path, (unsigned int)magic);
		return NULL;
	}
	fs_seek(f, 16, SEEK_CUR);
	fs_read(f, &width, 4);
	fs_read(f, &height, 4);
	dprintf("icon size: %dx%d\n", width, height);
	fs_seek(f, 2, SEEK_CUR);
	fs_read(f, &bpp, 2);
	dprintf("icon bpp: %d\n", bpp);
	ncolors = 1 << bpp;
	dprintf("icon colors: %d\n", ncolors);
	fs_seek(f, 16, SEEK_CUR);
	fs_read(f, &clr_used, 4);
	dprintf("icon colors used: %d\n", clr_used);
	fs_seek(f, 4, SEEK_CUR);

	for (i = 0; i < clr_used; i++) {
		fs_read(f, &b, 1);
		fs_read(f, &g, 1);
		fs_read(f, &r, 1);
		fs_seek(f, 1, SEEK_CUR);
		r >>= 4; g >>= 4; b >>= 4;
		*pal = (0xf << 12) | (r << 8) | (g << 4) | b;
		dprintf("pal[%d]: %04x\n", i, *pal);
		pal++;
	}

	size = height * width * bpp / 8;
	buf = malloc(size);
	rsize = fs_read(f, buf, size);
	dprintf("read %d of %d bytes\n", rsize, size);
	fs_close(f);
	return buf;

}

/* warning: issues pvr primitive font_header */
static void draw_text_box(text_box_t *tbox) {
	float tx, ty;
	char *tptr;

	if (!tbox) return;
	
	tptr = tbox->text;
	/* background */
	pvr_prim(&trans_header, sizeof(trans_header));
	draw_poly_box(tbox->x, tbox->y, tbox->w, tbox->h, tbox->z - 0.1f, tbox->a, 0.0f, 0.0f, 0.0f);
	draw_outline(tbox->x - tbox->t, tbox->y - tbox->t, tbox->w + tbox->t * 2, tbox->h + tbox->t * 2, tbox->z, tbox->t, 1.0f, tbox->r, tbox->g, tbox->b);

	pvr_prim(&font_header, sizeof(font_header));
	tx = tbox->x + tbox->xpad;
	ty = tbox->y + tbox->ypad;
	while (*tptr != '\0') {
		font_draw_strf(tx, ty, tbox->z + 0.1f, tbox->ta, tbox->tr, tbox->tg, tbox->tb, tptr);
		tptr += strlen(tptr) + 1;
		ty += tbox->yspacing;
	}
}

/* warning: issues pvr primitive font_header */
static void draw_menu(menu_t *menu) {
	float tx, ty;
	float bx, by;
	int i;

	if (!menu) return;

	menu_item_t *items = menu->items;
	pvr_prim(&trans_header, sizeof(trans_header));
	/* duplicated code from draw_text_box */
	draw_poly_box(menu->x, menu->y, menu->w, menu->h, menu->z - 0.1f, menu->a, 0.0f, 0.0f, 0.0f);
	draw_outline(menu->x - menu->t, menu->y - menu->t, menu->w + menu->t * 2, menu->h + menu->t * 2, menu->z, menu->t, 1.0f, menu->r, menu->g, menu->b);

	bx = menu->x + menu->xpad;
	by = menu->y + menu->ypad;
	tx = bx + menu->xpad_bar;
	ty = by + menu->ypad_bar;

	pvr_prim(&font_header, sizeof(font_header));
	font_draw_strf(tx, ty, menu->z + 0.1f, menu->ta, menu->tr, menu->tg, menu->tb, menu->text);
	ty += menu->yspacing * 1.5f;
	by += menu->yspacing * 1.5f;
	for (i = 0; i < menu->nitems; i++, items++) {
		/* status, activate, inc_dec not handled yet */
		if (i == menu->selected) {
			pvr_prim(&trans_header, sizeof(trans_header));
			draw_poly_box(bx, by, menu->w - menu->xpad * 2, menu->bar_h, menu->z - 0.1f, menu->ba, menu->br, menu->bg, menu->bb);
			pvr_prim(&font_header, sizeof(font_header));
		}
		font_draw_strf(tx, ty, menu->z + 0.1f, menu->ta, menu->tr, menu->tg, menu->tb, items->text);
		ty += menu->yspacing;
		by += menu->yspacing;
	}

}

static int *slot_list;
static int nslot_list;

static void menu_vmu_save(void) {
	int i, sel;
	text_box_sel_type.text = "Save >\0";
	cur_help_text_box = &text_box_help_save;
	csel = &selector_save;

	selector_save.box[1] = &text_box_sel_type;
	selector_save.min = 1;
	selector_save.sel[0] = -1;
	if (cur_slot != -1)
		selector_save.sel[1] = cur_slot - selector_save.min;
	else selector_save.sel[1] = 0;
	selector_save.sel[2] = -1;
	selector_save.active_sel = 1;
	selector_save.max = MAX_SLOTS;
	selector_save.hilite_list = slot_list;
	selector_save.nhilite = nslot_list;
	selector_save.disable = NONE;

	sel = edit_select_handle();
	if (sel >= 0) {
		sel += selector_save.min;
		if (save_levels(sel) != -1) {
			if (list_search(slot_list, sel, nslot_list) == -1) {
				/* Add the new level into the hilite list (if it's not already in there). */
				nslot_list++;
				slot_list = realloc(slot_list, nslot_list * sizeof(int));
				slot_list[nslot_list - 1] = sel;
				list_sort(slot_list, nslot_list);
				cur_slot = sel;
			}
			snd_sfx_play(SAMPLE_SUCCESS, 160, 128);
		} else {
			snd_sfx_play(SAMPLE_FAILURE, 160, 128);
		}
	} else {
	}
}

static void menu_vmu_load(void) {
	int i, sel;

	text_box_sel2_type.text = "Load >\0";
	cur_help_text_box = &text_box_help_load;
	csel = &selector_load;

	selector_load.box[2] = &text_box_sel2_type;
	selector_load.min = 1;
	selector_load.sel[0] = -1;
	selector_load.sel[1] = -1;
	if (cur_slot != -1)
		selector_load.sel[2] = cur_slot - selector_load.min;
	else selector_load.sel[2] = 0;
	selector_load.active_sel = 2;
	selector_load.max = MAX_SLOTS;
	selector_load.hilite_list = slot_list;
	selector_load.nhilite = nslot_list;
	selector_load.disable = NORMAL;

	sel = edit_select_handle();
	if (sel >= 0) {
		sel += selector_load.min;
		if (load_levels(sel) != -1) {
			/* We can't communicate back to the edit selector any other way. */
			/* This function could return a value, but it would be tricky to receive. */
			cur_slot = sel;
			selector_edit.sel[0] = 0;
			snd_sfx_play(SAMPLE_SUCCESS, 160, 128);
		} else {
			snd_sfx_play(SAMPLE_FAILURE, 160, 128);
		}
	}
}

static void menu_vmu_port(void) {
	if (scan_slots() >= 0) {
		snd_sfx_play(SAMPLE_SUCCESS, 160, 128);
	} else {
		snd_sfx_play(SAMPLE_FAILURE, 160, 128);
	}
}

static void menu_vmu_port_incdec(struct menu_t *menu, int item, int dir) {
	snd_sfx_play(SAMPLE_DISABLED, 160, 128);
}

int compar_int(const void *a, const void *b) {
	return *(int *)a - *(int *)b;
}

static int scan_slots(void) {
	file_t d;
	dirent_t *de;
	int i = 0;

	free(slot_list);
	slot_list = NULL;
	/* No current way to rewind a directory or obtain number of entries, portably */

	d = fs_open(LOAD_DIR, O_RDONLY | O_DIR);
	if (!d) return -1;
	while ((de = fs_readdir(d))) { 
		printf("entry: %s\n", de->name);
		i++;
	}
	fs_close(d);

	if (i == 0) {
		return 0;
	}
	slot_list = malloc(i * sizeof(int));
	printf("%d entries found, ", i);

	d = fs_open(LOAD_DIR, O_RDONLY | O_DIR);
	if (!d) return -1;
	i = 0;
	while ((de = fs_readdir(d))) { 
		if (strncmp(de->name, "brkout.", 7) == 0) {
			slot_list[i] = atoi(de->name + 7);
			i++;
		}
	}
	if (i == 0) {
		free(slot_list);
		slot_list = NULL;
		return 0;
	}
	nslot_list = i;
	printf("%d brkout slots", i);

	list_sort(slot_list, nslot_list);
	
	for (i = 0; i < nslot_list; i++) {
		printf("[%d]", slot_list[i]);
	}
	printf("\n");
	return nslot_list;
}

/* linear list search */
/* list must be monotonically increasing */
/* remove (list[i] > f) break condition to work with any list */
static int list_search(int *list, int f, int n) {
	int i;

	assert(list);
	for (i = 0; i < n; i++) {
		if (list[i] == f)
			return i;
		if (list[i] > f) 
			break;  
	}
	return -1;
}

/* search list for element >= f */
static int list_search_geq(int *list, int f, int n) {
	int i;
	assert(list);
	for (i = 0; i < n; i++) {
		if (list[i] >= f) 
			return i;
	}
	return -1;
}

static inline void list_sort(int *list, int n) {
	assert(list);
	assert(n > 0);
	qsort(list, n, sizeof(int), compar_int);
}

typedef struct {
	char *text1, *text2;
	int disable1, disable2;
} menu_item_seltype_data_t;

static menu_item_seltype_data_t menu_special_seltype_data[MENU_SPECIAL_ITEMS] = {
	{ "Copy >\0", "  To >\0", HILITE, HILITE },
	{ "Move >\0", "  To >\0", HILITE, NONE },
	{ "Swap >\0", "With >\0", HILITE, HILITE },
	{ NULL, NULL, NONE, NONE },
	{ NULL, NULL, HILITE, NONE },
	{ NULL, NULL, NONE, NONE },
	{ NULL, NULL, HILITE, NONE }
};

static void delete_level(int lvl) {
	if (lvl <= last_level) {
		free_level(levels[lvl]);
		memmove(&levels[lvl], &levels[lvl + 1], sizeof(level_t) * (last_level - lvl));
		last_level--;
	}
}

static void insert_level(int lvl) {
	if (last_level + 1 < NUM_LEVELS) {
		/* If we select the phantom level, memmove should do a 0-byte (no-op) copy. */
		memmove(&levels[lvl + 1], &levels[lvl], sizeof(level_t) * (last_level - lvl + 1));
		alloc_level(levels[lvl]);
		clear_level(levels[lvl]);
		last_level++;
	}
}

static void swap_level(int lvl1, int lvl2) {
	tmp_level.left = levels[lvl1].left;
	tmp_level.bricks = levels[lvl1].bricks;
	levels[lvl1].left = levels[lvl2].left;
	levels[lvl1].bricks = levels[lvl2].bricks;
	levels[lvl2].left = tmp_level.left;
	levels[lvl2].bricks = tmp_level.bricks;
}

static void menu_special_generic(void) {
	int sel1 = -1, sel2 = -1;
	int i;
	menu_item_seltype_data_t *msd = &menu_special_seltype_data[menu_special.selected];
	/* May be able to re-use selector_edit here. */
	
	text_box_sel_type.text  = msd->text1;
	text_box_sel2_type.text  = msd->text2;
	csel = &selector_special;

	selector_special.box[1] = &text_box_sel_type;
	selector_special.box[2] = &text_box_sel2_type;
	selector_special.min = selector_edit.min;
	selector_special.max = selector_edit.max;
	selector_special.sel[0] = -1;
	selector_special.sel[1] = selector_edit.sel[0];
	selector_special.sel[2] = selector_edit.sel[0];
	selector_special.top_row = selector_edit.top_row;
	selector_special.active_sel = 2;
	selector_special.hilite_list = selector_edit.hilite_list;
	selector_special.nhilite = selector_edit.nhilite;
	/* Selectability of phantom level, in this case. */
	selector_special.disable = msd->disable2;

	sel1 = selector_special.sel[1];
	sel1 += selector_special.min;

	/* Notice that I'm reimplementing the selector disable mechanism here,
	 * which is bad. */
	if ((sel1 == last_level + 1) && (msd->disable1 == HILITE)) {
		snd_sfx_play(SAMPLE_DISABLED, 160, 128);
		return;
	}

	/* An easy way to draw the level that's under the cursor is to make
	 * selector_edit.sel[0] track selector_special.sel[2].  edit_select_render
	 * would then draw the selected level.  However, there's no current way to
	 * specify that _handle update both variables. */

	if (msd->text1) {
		sel2 = edit_select_handle();
		if (sel2 < 0) return;
		sel2 += selector_special.min;
		if (sel1 == sel2) {
			snd_sfx_play(SAMPLE_DISABLED, 160, 128);
			return;
		}
	}

	switch(menu_special.selected) {
		case COPY:
			copy_level(levels[sel2], levels[sel1]);
			selector_edit.sel[0] = selector_special.sel[2];
			selector_edit.top_row = selector_special.top_row;
			break;
		case MOVE:
			/* The move semantics are confusing. */
			insert_level(sel2);
			if (sel2 < sel1) sel1++; /* adjust position for new level */
			swap_level(sel1, sel2);
			delete_level(sel1);
			if (sel1 < sel2) selector_special.sel[2]--;
			selector_edit.sel[0] = selector_special.sel[2];
			selector_edit.top_row = selector_special.top_row;
			break;
		case SWAP:
			swap_level(sel1, sel2);
			selector_edit.sel[0] = selector_special.sel[2];
			selector_edit.top_row = selector_special.top_row;
			break;
		case INSERT:
			insert_level(sel1);
			break;
		case DELETE:
			delete_level(sel1);
			break;
		case DELETE_ALL:
			dprintf("delete_all: last_level %d\n", last_level);
			for (i = 1; i <= last_level; i++) {
				dprintf("deleting level %d @ %08x\n", i, (uint32)levels[i].bricks);
				free_level(levels[i]);
			}
			last_level = 0;
			selector_edit.sel[0] = 0;
			break;
		case CLEAR:
			assert(sel1 < last_level + 1);
			clear_level(levels[sel1]);
			break;
		default:
			break;
	}

}

void draw_poly_tri(pvr_vertex_t *v1, pvr_vertex_t *v2, pvr_vertex_t *v3) {
	v1->flags = PVR_CMD_VERTEX;
	pvr_prim(v1, sizeof(pvr_vertex_t));
	v2->flags = PVR_CMD_VERTEX;
	pvr_prim(v2, sizeof(pvr_vertex_t));
	v3->flags = PVR_CMD_VERTEX_EOL;
	pvr_prim(v3, sizeof(pvr_vertex_t));
}

void draw_poly_tri_isosceles(float x, float y, float dx, float dy, float z, float a, float r, float g, float b) {
	pvr_vertex_t v1, v2, v3;
	assert(dx > 0); assert(dy != 0);

	v1.x = x; v2.x = v1.x + dx / 2; v3.x = v1.x + dx;
	v2.y = y; v1.y = v2.y + dy; v3.y = v1.y;
	v1.z = v2.z = v3.z = z;
	v1.argb = v2.argb = v3.argb = PVR_PACK_COLOR(a, r, g, b);
	v1.oargb = v2.oargb = v3.oargb = 0;
	v1.u = v2.u = v3.u = v1.v = v2.v = v3.v = 0.0f;
	/* Respect winding order (I think) */
	if (dy > 0) draw_poly_tri(&v1, &v2, &v3);
	else draw_poly_tri(&v3, &v2, &v1);
}


