#include "font.h"

/* initialize draw stuff: get our texture of the font, etc */
void font_init(void) {
	uint16 *vtex;
	int x, y;
	pvr_poly_cxt_t tmp_cxt;
	pvr_ptr_t txr_font;

	txr_font = pvr_mem_malloc(256*256*2);
	vtex = (uint16 *)txr_font;

	for (y = 0; y < 8; y++) {
		for (x = 0; x < 16; x++) {
			bfont_draw(vtex, 256, 1, y*16 + x);
			vtex += 16;
		}
		vtex += 23*256;
	}

    pvr_poly_cxt_txr(&tmp_cxt, PVR_LIST_TR_POLY, PVR_TXRFMT_ARGB4444 | PVR_TXRFMT_NONTWIDDLED,
	        256, 256, txr_font, PVR_FILTER_NONE);
    pvr_poly_compile(&font_header, &tmp_cxt);
	
}


/* Draws a font letter using two triangle strips */
void font_draw_char(int x1, int y1, int z1, float a, float r, float g, float b, int ch) {
	int	index, ix, iy;
	float	u1, v1, u2, v2;
	pvr_vertex_t	vert;

	if (ch == ' ') return;

	index = ch; // charmap[ch];
	/* FONT_CHAR_WIDTH is actual width of character in texture */
	ix = (index % FONT_PERROW) * FONT_CHAR_WIDTH;
	iy = (index / FONT_PERROW) * FONT_CHAR_HEIGHT;
	u1 = ix * 1.0f / FONT_PIC_WIDTH;
	v1 = iy * 1.0f / FONT_PIC_HEIGHT;
	u2 = (ix+FONT_WIDTH) * 1.0f / FONT_PIC_WIDTH;
	v2 = (iy+FONT_HEIGHT) * 1.0f / FONT_PIC_HEIGHT;

	vert.flags = PVR_CMD_VERTEX;
	vert.x = x1;
	vert.y = y1 + FONT_HEIGHT*FONT_DOUBLE;
	vert.z = z1;
	vert.u = u1;
	vert.v = v2;

	vert.argb = PVR_PACK_COLOR(a, r, g, b);
	vert.oargb = 0;
	pvr_prim(&vert, sizeof(vert));
	
	vert.x = x1;
	vert.y = y1;
	vert.u = u1;
	vert.v = v1;
	pvr_prim(&vert, sizeof(vert));
	
	vert.x = x1 + FONT_WIDTH*FONT_DOUBLE;
	vert.y = y1 + FONT_HEIGHT*FONT_DOUBLE;
	vert.u = u2;
	vert.v = v2;
	pvr_prim(&vert, sizeof(vert));

	vert.flags = PVR_CMD_VERTEX_EOL;
	vert.x = x1 + FONT_WIDTH*FONT_DOUBLE;
	vert.y = y1;
	vert.u = u2;
	vert.v = v1;
	pvr_prim(&vert, sizeof(vert));
}


/* Draw a set of textured polygons at the given depth and color that
   represent a string of text. */
void font_draw_strf(float x1, float y1, float z1, float a, float r, float g, float b, char *fmt, ...) {
	va_list args;
	char *s;
	static char strbuf[1024];
	
	va_start(args, fmt);
	vsprintf(strbuf, fmt, args);
	va_end(args);

	s = strbuf;
	while (*s) {
		font_draw_char(x1, y1, z1, a, r, g, b, *s++);
		x1+=FONT_WIDTH * FONT_DOUBLE + 1;
	}
}

